/******************************************************************************


Copyright 1993, 1998  The Open Group

Permission to use, copy, modify, distribute, and sell this software and its
documentation for any purpose is hereby granted without fee, provided that
the above copyright notice appear in all copies and that both that
copyright notice and this permission notice appear in supporting
documentation.

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
OPEN GROUP BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of The Open Group shall not be
used in advertising or otherwise to promote the sale, use or other dealings
in this Software without prior written authorization from The Open Group.

Author: Ralph Mor, X Consortium
******************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <X11/ICE/ICElib.h>
#include "ICElibint.h"
#include <X11/ICE/ICEutil.h>

#include <time.h>
#define Time_t time_t

#ifdef HAVE_LIBBSD
#include <bsd/stdlib.h>	/* for arc4random_buf() */
#endif

static int was_called_state;

#ifndef HAVE_ARC4RANDOM_BUF

static void
emulate_getrandom_buf (
	char *auth,
	int len
)
{
    long    ldata[2];
    int	    seed;
    int	    value;
    int	    i;

#ifdef ITIMER_REAL
    {
	struct timeval  now;
	X_GETTIMEOFDAY (&now);
	ldata[0] = now.tv_sec;
	ldata[1] = now.tv_usec;
    }
#else /* ITIMER_REAL */
    {
	long    time ();
	ldata[0] = time ((long *) 0);
	ldata[1] = getpid ();
    }
#endif /* ITIMER_REAL */
    seed = (ldata[0]) + (ldata[1] << 16);
    srand (seed);
    for (i = 0; i < len; i++)
    {
	value = rand ();
	auth[i] = value & 0xff;
    }
}

#ifndef _GNU_SOURCE
#define _GNU_SOURCE // needed on SLE11 for O_CLOEXEC
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <linux/random.h>
#include <sys/syscall.h>

int getentropy_urandom(void *buffer, size_t length)
{
	int random_fd = -1;
	ssize_t res = -1;
	size_t filled = 0;

	if( length > 256 )
	{
		errno = EIO;
		return -1;
	}

	random_fd = open("/dev/urandom", O_RDONLY | O_CLOEXEC);

	if( random_fd == -1 )
	{
		return -1;
	}

	while( filled < length )
	{
		res = read(random_fd, (char*)buffer + filled, length - filled);

		if( res == -1 )
		{
			// shouldn't actually happen acc. to man(4) random,
			// but you never know
			if( errno == EINTR )
				continue;

			close(random_fd);
			return -1;
		}
		else if( res == 0 )
		{
			// no more bytes available? should not happen
			errno = EIO;
			close(random_fd);
			return -1;
		}

		filled += res;
	}

	close(random_fd);
	return 0;
}

int getentropy_getrandom(void *buffer, size_t length)
{
	int res;
	size_t filled = 0;

	if( length > 256 )
	{
		errno = EIO;
		return -1;
	}

	while( filled < length )
	{
#ifdef SYS_getrandom
		/*
		 * glibc does not contain a syscall wrapper for this in older
		 * versions
		 */
		res = syscall(SYS_getrandom, (char*)buffer + filled, length - filled, 0);
#else
#	warning no getrandom
		errno = ENOSYS;
		return -1;
#endif // SYS_getrandom

		if( res == -1 )
		{
			if( errno == EINTR )
				continue;

			return -1;
		}
		else if( res == 0 )
		{
			// no more bytes available? should not happen
			errno = EIO;
			return -1;
		}

		filled += res;
	}

	return 0;
}

int getentropy_emulate(void *buffer, size_t length)
{
	/*
	 * check at runtime whether we have a getrandom system call available,
	 * otherwise fall back to urandom approach. autoconf check for
	 * getrandom() does not work, because there's been no declaration for
	 * it for years.
	 */
	int res = getentropy_getrandom(buffer, length);

	if( res == -1 && errno == ENOSYS )
	{
		return getentropy_urandom(buffer, length);
	}

	return res;
}

static void
arc4random_buf (
	char *auth,
	int len
)
{
    int	    ret;

#if HAVE_GETENTROPY
    /* weak emulation of arc4random through the entropy libc */
    ret = getentropy (auth, len);
#else
    ret = getentropy_emulate (auth, len);
#endif /* HAVE_GETENTROPY */
    if (ret == 0)
	return;

    emulate_getrandom_buf (auth, len);
}

#endif /* !defined(HAVE_ARC4RANDOM_BUF) */

/*
 * MIT-MAGIC-COOKIE-1 is a sample authentication method implemented by
 * the SI.  It is not part of standard ICElib.
 */


char *
IceGenerateMagicCookie (
	int len
)
{
    char    *auth;

    if ((auth = malloc (len + 1)) == NULL)
	return (NULL);

    arc4random_buf (auth, len);

    auth[len] = '\0';
    return (auth);
}



IcePoAuthStatus
_IcePoMagicCookie1Proc (
	IceConn		iceConn,
	IcePointer	*authStatePtr,
	Bool 		cleanUp,
	Bool		swap,
	int     	authDataLen,
	IcePointer	authData,
	int 		*replyDataLenRet,
	IcePointer	*replyDataRet,
	char    	**errorStringRet
)
{
    if (cleanUp)
    {
	/*
	 * We didn't allocate any state.  We're done.
	 */

	return (IcePoAuthDoneCleanup);
    }

    *errorStringRet = NULL;

    if (*authStatePtr == NULL)
    {
	/*
	 * This is the first time we're being called.  Search the
	 * authentication data for the first occurence of
	 * MIT-MAGIC-COOKIE-1 that matches iceConn->connection_string.
	 */

	unsigned short  length;
	char		*data;

	_IceGetPoAuthData ("ICE", iceConn->connection_string,
	    "MIT-MAGIC-COOKIE-1", &length, &data);

	if (!data)
	{
	    const char *tempstr =
		"Could not find correct MIT-MAGIC-COOKIE-1 authentication";

	    *errorStringRet = strdup(tempstr);

	    return (IcePoAuthFailed);
	}
	else
	{
	    *authStatePtr = (IcePointer) &was_called_state;

	    *replyDataLenRet = length;
	    *replyDataRet = data;

	    return (IcePoAuthHaveReply);
	}
    }
    else
    {
	/*
	 * We should never get here for MIT-MAGIC-COOKIE-1 since it is
	 * a single pass authentication method.
	 */

	const char *tempstr =
	    "MIT-MAGIC-COOKIE-1 authentication internal error";

	*errorStringRet = strdup(tempstr);

	return (IcePoAuthFailed);
    }
}

IcePoAuthProc	_IcePoAuthProcs[] = {_IcePoMagicCookie1Proc};


IcePaAuthStatus
_IcePaMagicCookie1Proc (
	IceConn		iceConn,
	IcePointer	*authStatePtr,
	Bool		swap,
	int     	authDataLen,
	IcePointer	authData,
	int 		*replyDataLenRet,
	IcePointer	*replyDataRet,
	char    	**errorStringRet
)
{
    *errorStringRet = NULL;
    *replyDataLenRet = 0;
    *replyDataRet = NULL;

    if (*authStatePtr == NULL)
    {
	/*
	 * This is the first time we're being called.  We don't have
	 * any data to pass to the other client.
	 */

	*authStatePtr = (IcePointer) &was_called_state;

	return (IcePaAuthContinue);
    }
    else
    {
	/*
	 * Search the authentication data for the first occurence of
	 * MIT-MAGIC-COOKIE-1 that matches iceConn->connection_string.
	 */

	unsigned short  length;
	char		*data;

	_IceGetPaAuthData ("ICE", iceConn->connection_string,
	    "MIT-MAGIC-COOKIE-1", &length, &data);

	if (data)
	{
	    IcePaAuthStatus stat;

	    if (authDataLen == length &&
	        memcmp (authData, data, authDataLen) == 0)
	    {
		stat = IcePaAuthAccepted;
	    }
	    else
	    {
		const char *tempstr
		    = "MIT-MAGIC-COOKIE-1 authentication rejected";

		*errorStringRet = strdup(tempstr);

		stat = IcePaAuthRejected;
	    }

	    free (data);
	    return (stat);
	}
	else
	{
	    /*
	     * We should never get here because in the ConnectionReply
	     * we should have passed all the valid methods.  So we should
	     * always find a valid entry.
	     */

	    const char *tempstr =
		"MIT-MAGIC-COOKIE-1 authentication internal error";

	    *errorStringRet = strdup(tempstr);

	    return (IcePaAuthFailed);
	}
    }
}

IcePaAuthProc	_IcePaAuthProcs[] = {_IcePaMagicCookie1Proc};
